; Program ...: Chgbits.ASM
; Author ....: Kenneth N. Getz
; Date ......: October 1, 1987
; Version ...: dBASE III PLUS (or The Developer's Release)
;
;*****************************************************************************
CODE    SEGMENT BYTE PUBLIC     'CODE'  
;-----------------------------------------------------------------------------
SAVE    MACRO   R1,R2,R3,R4,R5,R6,R7,R8
;
; Macro to save registers passed in a list, separated by commas.
;
        IRP     X,<R1,R2,R3,R4,R5,R6,R7,R8>
                IFNB    <X>
                        PUSH X  
                ENDIF
        ENDM
        ENDM    
;-----------------------------------------------------------------------------
RESTORE MACRO   R1,R2,R3,R4,R5,R6,R7,R8
;
; Macro to restore registers passed in a list, separated by commas.
; Registers are popped in the reverse order from which they are passed.
;
        IRP     X,<R8,R7,R6,R5,R4,R3,R2,R1>
                IFNB    <X>
                        POP X
                ENDIF
        ENDM
        ENDM    
;-----------------------------------------------------------------------------
MOVSEG  MACRO   R1,R2
;
; Macro to move segment value from the second parameter into the first.
; To the programmer, works like MOV.
;
        PUSH    R2
        POP     R1
        ENDM
;-----------------------------------------------------------------------------
GETNUM  MACRO   REG
;
; Calls Getval to get a value and places it into the register passed in.
;
        CALL    GETVAL
        MOV     REG,AL
        ENDM
;-----------------------------------------------------------------------------
NEXTDIGIT16     MACRO
LOCAL   SPACE1, SPACE2, SPACE3, SPACE4
;
; Macro to skip "white space" and find the next hex digit in a string.
;
        CMP     BYTE PTR [SI], 0       ; See if we got a null to start with.
        JE      SPACE4                 ; If so, we're done.
SPACE1: CMP     BYTE PTR [SI],'0'      ; Compare to ASCII 0.
        JGE     SPACE2                 ; If so, keep checking.
        INC     SI                     ; If not, move on to next character.
        JMP     SPACE1                 ; Go back for more.
SPACE2: CMP     BYTE PTR [SI],'F'      ; Compare to ASCII F.
        JLE     SPACE3                 ; If less, we're still OK.
        INC     SI                     ; Otherwise, move on to next character.
        JMP     SPACE1                 ; Go back for more.
SPACE3: CMP     BYTE PTR [SI],'9'      ; Compare to ASCII 9.
        JLE     SPACE4                 ; If it's less, digit's OK.
        CMP     BYTE PTR [SI],'A'      ; Compare to A.
        JGE     SPACE4                 ; If greater, digit's OK.
        INC     SI                     ; Otherwise, check next character.
        JMP     SPACE1                 ; Go back for more.
SPACE4: NOP
        ENDM
;-----------------------------------------------------------------------------
NEXTDIGIT10     MACRO
        LOCAL   SPACE1, SPACE2, SPACE3
;
; Macro to skip "white space" and find the next decimal digit in a string.
;
        CMP     BYTE PTR [SI], 0       ; See if we got a null to start with.
        JE      SPACE3                 ; If so, we're done.
SPACE1: CMP     BYTE PTR [SI],'0'      ; Compare to ASCII 0.
        JGE     SPACE2                 ; If so, keep checking.
        INC     SI                     ; If not, move on to next character.
        JMP     SPACE1                 ; Go back for more.
SPACE2: CMP     BYTE PTR [SI],'9'      ; Compare to ASCII 9.
        JLE     SPACE3                 ; If it's less, digit's OK.
        INC     SI                     ; Otherwise, check next character.
        JMP     SPACE1                 ; Go back for more.
SPACE3: NOP
        ENDM
;-----------------------------------------------------------------------------
SIZE    EQU     2                      ; Defines how many bytes can be Put
                                       ;    at one time.
;-----------------------------------------------------------------------------
MAIN    PROC    FAR
        ASSUME  CS:CODE

START:  JMP     ENTRY
        DB      13,'Chgbits.BIN',10,13
        DB      'Written by Kenneth N. Getz.',10,13 
        DB      'Copyright 1987, Ashton-Tate',10,13
        DB      26
BUFFER  DB      8 DUP(?)               ; Workspace for binary conversions.
LEN     DW      0                      ; Length of the parameter string.
PARAMS  DB      255 DUP(?)             ; Local copy of the parameter string.
SEGSTR  DB      4 DUP(?)               ; String version of segment.
OFFSTR  DB      4 DUP(?)               ; String version of offset.
SEGNUM  DW      0                      ; Segment as an integer.
POKE    DB      SIZE DUP (?)           ; Bytes to poke in memory.
NUM     DW      0
;
; Lookup table based on offsets into jump table.  Uses parameter passed
; in to jump to correct procedure.
;
;               A B C D E F G H I J K L M N O P Q R S T U V W X Y Z
LOOKUP  DB      3,8,0,7,0,0,1,0,0,0,0,0,0,5,4,2,0,0,0,0,0,0,0,6,0,0

TABLE   LABEL   WORD
        DW      DOPEEK
        DW      DOPOKE
        DW      DOAND
        DW      DO_OR
        DW      DONOT
        DW      DOXOR
        DW      DODEC
        DW      DOBIN
        
ENTRY:  SAVE    AX,BX,CX,DX,DI,SI,BP
        PUSH    DS                     ; Save dBASE's DS.
        XOR     CX,CX                  ; Clear out CX.
        MOV     CL, BYTE PTR [BX-1]    ; Get the length byte.
        INC     CL                     ; Add one for the null byte.
        MOV     LEN,CX                 ; Store it away.

        CALL    GETPARS                ; Get the parameter string.
        MOVSEG  DS,CS                  ; Make DS = CS.
        CALL    GETOP                  ; Get procedure address or set CF.
        JC      BADOP                  ; If CF set, we had an error.
        MOV     BP,AX                  ; Get set for jump to procedure.
        CALL    TABLE[BP]              ; Call the requested operation.
        POP     DS                     ; Restore dBASE's DS.
        CMP     BP,14                  ; If we're doing a binary translation,
        JE      BINOUT                 ;    finish up the operation.
        CALL    DEC                    ; Otherwise, finish decimal operations.
        JMP     DONE
BINOUT: CALL    BIN                    ; Finish binary operation.
        JMP     DONE
BADOP:  POP     DS                     ; On error -- reset dBASE's DS.
        MOV     BYTE PTR [BX],'X'      ; Return an "X" to indicate error.
        MOV     BYTE PTR [BX+1],0      ; Place null byte to end string.
DONE:   RESTORE AX,BX,CX,DX,DI,SI,BP
        RET
MAIN    ENDP
;-----------------------------------------------------------------------------
; Convert binary string to decimal number.  This routine just converts
; the string to an integer in DL.  The procedure DEC converts it to a 
; string representation later.
;
DODEC   PROC    NEAR
        ; SI comes in pointing to the next hex digit after the op code.
        MOV     CX,8                   ; Maximum length is 8.
        XOR     DX,DX                  ; Clear out DX.
DECLOOP:
        LODSB                          ; Get a byte in AL.
        SUB     AL,'0'                 ; Get offset from '0'.
        SHL     DL,1                   ; Multiply DL by 2.
        ADD     DL,AL                  ; Add either a 0 or a 1 to DL.
        LOOP    DECLOOP                ; Go back for more.
        RET
DODEC   ENDP
;-----------------------------------------------------------------------------
; Convert decimal number to binary.  We do this by shifting the bits left.
; If the carry flag is set, we had a 1. Otherwise, we had a 0.
;
DOBIN   PROC    NEAR
        SAVE    AX,BX,CX,DX,SI,DI
        CALL    GETPOKE                ; Get the decimal information.
        MOV     DL,BYTE PTR POKE       ; Point DL to poke address.
        MOV     CX,8                   ; Maximum length is 8.
        MOV     DI,OFFSET BUFFER       ; Point DI to buffer.
SHIFTLOOP:
        MOV     AL,'0'                 ; Assume we had a zero.
        SHL     DL,1                   ; Shift the number left 1.
        JNC     NOCARRY                ; If there's no carry, we had a 0.
        MOV     AL,'1'                 ; Otherwise, we had a 1.
NOCARRY:
        STOSB                          ; Store the digit ('0' or '1') in AL.
        LOOP    SHIFTLOOP              ; Go back for more.
        RESTORE AX,BX,CX,DX,SI,DI
        RET
DOBIN   ENDP
;-----------------------------------------------------------------------------
; Copies the binary string buffer out to dBASE.  
;
BIN     PROC    NEAR
        SAVE    DS,ES                  ; Save the segment registers.
        MOVSEG  ES,DS                  ; Make ES = DS (for DI).
        MOVSEG  DS,CS                  ; Make DS = CS (for SI).
        MOV     DI,BX                  ; Point DI to dBASE's variable.
        MOV     SI,OFFSET BUFFER       ; Point SI to string to copy.
        MOV     CX,8                   ; Maximum length is 8.
        REP     MOVSB                  ; Copy out the string.
        MOV     AL,0                   ; Add a null byte at the end.
        STOSB
        RESTORE DS,ES                  ; Restore segment registers.
        RET
BIN     ENDP
;-----------------------------------------------------------------------------
; Get a single byte from a particular memory address and put it in DL.
;
DOPEEK  PROC    NEAR
        CALL    GETDATA                ; Get the dBASE parameters.
        CALL    SETPOS                 ; Set correct segment and offset.
        SUB     DH,DH                  ; Clear out DH.
        MOV     DL,ES:[DI]             ; Get the peek byte.
        RET
DOPEEK  ENDP
;-----------------------------------------------------------------------------
; Put SIZE number of bytes at a chosen memory location.
;
DOPOKE  PROC    NEAR
        CALL    GETDATA                ; Get the dBASE parameters.
        CALL    GETPOKE                ; Get the values to Put.
        CALL    SETPOS                 ; Set the correct segment and offset.
        CALL    PUTPOKE                ; Put the values in memory there.
        RET
DOPOKE  ENDP
;-----------------------------------------------------------------------------
; Logically AND two bytes and return answer in DL.
;
DOAND   PROC    NEAR
        GETNUM  DH                     ; Get a value into DH.
        GETNUM  DL                     ; Get another into DL.
        AND     DL,DH                  ; AND them into DL.
        SUB     DH,DH                  ; Clear out DH.
        RET
DOAND   ENDP
;-----------------------------------------------------------------------------
; Logically OR two bytes and return answer in DL.
;
DO_OR   PROC    NEAR
        GETNUM  DH                     ; Get a value into DH.
        GETNUM  DL                     ; Get another into DL.
        OR      DL,DH                  ; OR them into DL.
        SUB     DH,DH                  ; Clear out DH.
        RET             
DO_OR   ENDP
;-----------------------------------------------------------------------------
; Logically XOR two bytes and return answer in DL.
;
DOXOR   PROC    NEAR
        GETNUM  DH                     ; Get a value into DH.
        GETNUM  DL                     ; Get another into DL.
        XOR     DL,DH                  ; XOR them into DL.
        SUB     DH,DH                  ; Clear out DH.
        RET
DOXOR   ENDP
;-----------------------------------------------------------------------------
; Logically NOT the byte stored in DL.
;
DONOT   PROC    NEAR                   
        GETNUM  DL                     ; Get a value into DL.
        NOT     DL                     ; NOT it.
        SUB     DH,DH                  ; Clear out DH.
        RET
DONOT   ENDP                                    
;-----------------------------------------------------------------------------
; Pokes the values stored in POKE into RAM at the indicated SEGMENT:OFFSET.
; CS:SI points to POKE, ES:DI points to RAM location at which to poke.
;
PUTPOKE PROC    NEAR
        MOV     SI,OFFSET POKE         ; Point SI to POKE.
        MOV     DX,NUM                 ; Sets dBASE return value in decimal.
        MOV     CX,DX                  ; Get max length in CX.
        REP     MOVSB                  ; Store the bytes.
        RET
PUTPOKE ENDP
;-----------------------------------------------------------------------------
; Given the segment and offset strings passed in from dBASE, this procedure
; converts them to integers and points ES to the segment and DI to the 
; offset.
;
SETPOS  PROC    NEAR
        SAVE    SI
        MOV     SI,OFFSET SEGSTR       ; Get address of segment string.
        CALL    CONVERT                ; Convert it to binary in CX.
        MOV     SEGNUM,CX              ; Store it away.
        MOV     SI,OFFSET OFFSTR       ; Get address of offset string.
        CALL    CONVERT                ; Convert it to binary in CX.
        MOV     DI,CX                  ; Move it to DI.
        MOV     ES,SEGNUM              ; Set ES to correct segment.
        RESTORE SI
        RET
SETPOS  ENDP

;-----------------------------------------------------------------------------
; Converts the 4-digit hexadecimal string pointed to by CS:SI into a 
; binary word in CX.
;
CONVERT PROC    NEAR
        CALL    CVT                    ; Convert the first two digits.
        MOV     AH,AL                  ; Move the result to upper byte.
        AND     AL,0                   ; Clear out lower byte.
        MOV     CX,AX                  ; Move into CX.
        CALL    CVT                    ; Get lower two bytes.
        ADD     CX,AX                  ; Add to what's already in CX.
        RET                            
CONVERT ENDP
;-----------------------------------------------------------------------------
; Convert two hex digits to binary.  Assumes SI points to string, returns
; value in AX.
;
CVT     PROC    NEAR
        SAVE    BX,CX,DX
        MOV     BX,3030H               ; Store the conversion bias.
        LODSB                          ; Get a byte.
        CMP     AL,'9'                 ; See if it's less than A.
        JBE     NEXT1                  ; If so, do nothing.
        ADD     BH,7                   ; Otherwise, add bias for hex digit.
NEXT1:  MOV     AH,AL                  ; Shift byte over in word.
        LODSB                          ; Get next byte.
        CMP     AL,'9'                 ; Again see if it's less than A.
        JBE     NEXT2                  ; If so, do nothing.
        ADD     BL,7                   ; Add to bias for hex digit.
NEXT2:  SUB     AX,BX                  ; Subtract off bias.
        MOV     DL,AH                  ; Move high-order byte to DL.
        MOV     CL,4                   ; Set up for shift.
        SHL     DL,CL                  ; Multiply by 16.
        ADD     AL,DL                  ; Add low-order byte.
        SUB     AH,AH                  ; Clear out AH.
        RESTORE BX,CX,DX
        RET
CVT     ENDP
;-----------------------------------------------------------------------------
; Get the offset and segment.  SI ends up after last byte of offset 
; (either on poke byte or null, depending on where it's called from.)
; Expects SI to point to PARAMS when it is called.
;
GETDATA PROC    NEAR
        SAVE    DI,BX,CX
        MOV     CX,4                   ; Set CX for copy loop.
        MOV     DI,OFFSET SEGSTR       ; Point DI to SEGMENT storage.
        REP     MOVSB                  ; Copy from dBASE to storage.
        INC     SI                     ; Skip over colon.
        MOV     DI,OFFSET OFFSTR       ; Point DI to storage.
        MOV     CX,4                   ; Set CX for copy loop.
        REP     MOVSB                  ; Copy from dBASE to storage.
        RESTORE DI,BX,CX
        RET
GETDATA ENDP
;-----------------------------------------------------------------------------
; Get the option letter, find the correct routine to run if any, move
; the pointer along in the option list to the next non-whitespace
; character, and set the Carry Flag if an error occurred in the parameter
; list.
;
GETOP   PROC    NEAR
        PUSH    BX                     
        MOV     SI,OFFSET PARAMS       ; Point SI to the parameter string.
        LODSB                          ; Move option letter into AL.
        NEXTDIGIT16                    ; Skip white space.
        SUB     AH,AH                  ; Clear out AH.
        SUB     AL,'A'                 ; Get offset from 'A'.
        CMP     AL,0                   ; See if it was less than 'A'.
        JL      NO_OP                  ; If so, we have an error.
        CMP     AL,25                  ; Compare to 'Z'.
        JG      NO_OP                  ; If greater, we have an error.
        MOV     BX,OFFSET LOOKUP       ; Set up for lookup.
        XLAT                           ; Translate and find correct routine.
        CMP     AL,0                   ; See if XLAT returned a valid number.
        JE      NO_OP                  ; If not, we have an error.
        DEC     AX                     ; Otherwise, bias to 0.
        SHL     AX,1                   ; Multiply by 2 for word offsets.
        CLC                            ; Make sure Carry Flag is not set.
        JMP     ENDOP                  ; Get out.
NO_OP:  STC                            ; Error -- set the Carry Flag.
ENDOP:  POP     BX
        RET
GETOP   ENDP
;-----------------------------------------------------------------------------
; Clear out the poke storage space for multiple calls to the procedure and 
; gets a copy of the entire dBASE parameter string.
;
GETPARS PROC    NEAR
        SAVE    AX, SI, DI, CX         ; Save the environment.
        MOV     CX,255                 ; Set up counter for loop.
        MOV     DI, OFFSET PARAMS      ; Get the address to write to.
        AND     AL,0                   ; Clear out AL.
        REP     STOSB                  ; Puts 255 zeros in to clear string.
        MOV     CX,255                 ; Set counter back to 255.
        MOV     DI, OFFSET PARAMS      ; Set DI back to beginning of storage.
        MOV     SI,BX                  ; Point SI to dBASE's parameter string.
PARLOOP:
        LODSB                          ; Get a byte.
        CMP     AL,0                   ; See if it's a null.
        JE      PARDONE                ; If so, we're done.
        CMP     AL,'a'                 ; See if below 'a'.
        JB      NOTALPHA               ; If so, leave it alone.
        CMP     AL,'z'                 ; See if above 'z'.
        JA      NOTALPHA               ; If so, leave it alone.
        AND     AL,0DFH                ; Otherwise, capitalize it.
NOTALPHA:       
        STOSB                          ; In any case, copy the byte.
        LOOP    PARLOOP                ; And loop back for more.
PARDONE:                        
        RESTORE AX, SI, DI, CX         ; Restore environment.
        RET                            
GETPARS ENDP
;-----------------------------------------------------------------------------
; Clear the poke byte storage area for multiple calls to the routine.
;
CLEARPK PROC    NEAR
        SAVE    AX,CX
        MOV     CX,SIZE                ; Set up for SIZE words to clear.
        XOR     AX,AX                  ; We're clearing with 0.
        MOV     DI,OFFSET POKE         ; Point DI to storage.
        REP     STOSB                  ; Do the clearing.
        RESTORE AX,CX                  
        RET
CLEARPK ENDP
;-----------------------------------------------------------------------------
; Get the poke bytes.  Will read up to SIZE bytes into temporary storage
; for our program's use.  Stores in NUM the actual number of bytes read.
;
GETPOKE PROC    NEAR
        SAVE    AX                    
        MOV     CX,SIZE                ; Set up for copy loop.
        MOV     DI,OFFSET POKE         ; Point DI to storage.
GETLOOP:                               
        CLC                            ; Clear Carry Flag.
        CALL    GETVAL                 ; Get a byte from dBASE.
        JC      GETDONE                ; Carry set if NULL byte found.
        STOSB                          ; Otherwise, store it.
        LOOP    GETLOOP                ; Loop back (SIZE times) for more.
GETDONE:
        MOV     NUM,SIZE               ; Store away maximum number of bytes.
        SUB     NUM,CX                 ; Subtract extra leaving number read.
        RESTORE AX
        RET
GETPOKE ENDP
;-----------------------------------------------------------------------------
; Gets decimal string parameter pointed to by SI and puts the value into 
; AL.  Skips anything except numbers, and ends reading when it gets
; to a non-integer.
;
GETVAL  PROC    NEAR
        SAVE    CX
        MOV     CL,10                  ; Set up for base 10 multiplies.
        XOR     AX,AX                  ; Clear out AX.
        CMP     BYTE PTR [SI],0        ; See if we're on a null.
        JNE     OKVAL                  ; If not, we're OK. 
        STC                            ; Otherwise, set the carry flag.
        JMP     GETOUT                 ; Get out.
OKVAL:  NEXTDIGIT10                    ; Go to the non-whitespace char.
INLOOP: MOV     DL, BYTE PTR [SI]      ; Move it into DL.
        CMP     DL,'0'                 ; Compare it to '0'.
        JL      @LOOP                  ; Less?  Not a digit.
        CMP     DL,'9'                 ; Compare to '9'.
        JG      @LOOP                  ; Greater?  Not a digit.
        SUB     DL,'0'                 ; Otherwise, subtract bias.
        MUL     CL                     ; Multiply by 10.
        ADD     AL,DL                  ; Add into accumulator.
        INC     SI                     ; Go on to next digit.
        JMP     INLOOP                 ; Loop back.
@LOOP:  CLC                            ; No error -- clear carry flag.
GETOUT: RESTORE CX
        RET
GETVAL  ENDP
;-----------------------------------------------------------------------------
; Converts a binary word in DX into a decimal string representation.
;
DEC     PROC    NEAR
        SAVE    AX,CX,DX,SI,DI
        MOV     AX,DX                  ; Get the original number.
        MOV     SI,10                  ; Set SI up for divides.
        XOR     CX,CX                  ; Clear out CX.
NON_ZERO:
        XOR     DX,DX                  ; Clear out DX.
        DIV     SI                     ; Divide AX by 10.
        PUSH    DX                     ; Save DX.
        INC     CX                     ; CX = CX + 1.
        OR      AX,AX                  ; See if AX = 0.
        JNE     NON_ZERO               ; If not, do it again.
        XOR     DI,DI                  ; Clear out DI.
WRITE_DIGIT_LOOP:
        POP     DX                     ; Pop a digit.
        ADD     DL,'0'                 ; Add ASCII offset.
        MOV     BYTE PTR [BX+DI],DL    ; Store it away.
        INC     DI                     ; Go on to the next digit.
        LOOP    WRITE_DIGIT_LOOP       ; Go back for more.
CLEAR_LOOP:
        CMP     DI,LEN                 ; See if we're done.
        JE      END_DECIMAL            ; If so, jump out.
        MOV     BYTE PTR[BX+DI],0      ; Otherwise, pad with a null.
        INC     DI                     ; Go to next spot.
        JMP     CLEAR_LOOP             ; Jump back for more.
END_DECIMAL:
        RESTORE AX,CX,DX,SI,DI
        RET
DEC     ENDP
;-----------------------------------------------------------------------------
;
CODE    ENDS
;*****************************************************************************
        END     START                           

        

